import os
import sys
import time
import requests
import unittest
from typing import List

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from language_pipes.cli import main
from language_pipes.util.chat import ChatMessage, ChatRole

MODEL = "Qwen/Qwen3-1.7B"
# MODEL = "meta-llama/Llama-3.2-1B-Instruct"

def start_node(node_id: str, max_memory: float, peer_port: int, job_port: int, oai_port: int = None, bootstrap_port: int = None):
    args = ["run", 
        "--node-id", node_id, 
        "--hosted-models", f"{MODEL}::cpu::{max_memory}", 
        "--peer-port", str(peer_port),
        "--job-port", str(job_port)
    ]
    if oai_port is not None:
        args.extend(["--oai-port", str(oai_port)])
    
    if bootstrap_port is not None:
        args.extend(["--bootstrap-address", "localhost", "--bootstrap-port", str(bootstrap_port)])

    return main(args)

def oai_complete(port: int, messages: List[ChatMessage], retries: int = 0):
    try:
        res = requests.post(f"http://localhost:{port}/v1/chat/completions", json={
            "model": MODEL,
            "max_completion_tokens": 10,
            "messages": [m.to_json() for m in messages]
        })
        if res.status_code != 200:
            raise Exception(f"Failed to complete: {res.text}")
        return res.json()
    except Exception as e:
        print(e)
        if retries < 5:
            time.sleep(5)
            return oai_complete(port, messages, retries + 1)


class OpenAITests(unittest.TestCase):
    def test_single_node(self):
        start_node("node-1", 5, 5000, 5050, 6000)
        res = oai_complete(6000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res["choices"][0]["message"]["content"] + "\"")
        self.assertTrue(len(res["choices"]) > 0)

    def test_400_codes(self):
        start_node("node-1", 5, 5000, 5050, 6000)
        messages = [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ]
        res = requests.post("http://localhost:6000/v1/chat/completions", json={
            "messages": [m.to_json() for m in messages]
        })

        self.assertEqual(400, res.status_code)

        res = requests.post("http://localhost:6000/v1/chat/completions", json={
            "model": MODEL
        })

        self.assertEqual(400, res.status_code)

        res = requests.post("http://localhost:6000/v1/chat/completions", json={
            "model": MODEL,
            "messages": []
        })

        self.assertEqual(400, res.status_code)

    def test_double_node(self):
        start_node("node-1", 2, 5000, 5050, 6000)
        time.sleep(10)
        start_node("node-2", 3, 5001, 5051, None, 5000)
        time.sleep(10)
        res = oai_complete(6000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res["choices"][0]["message"]["content"] + "\"")
        self.assertTrue(len(res["choices"]) > 0)

    def test_triple_node(self):
        start_node("node-1", 1, 5000, 5050, 6000)
        time.sleep(10)
        start_node("node-2", 1, 5001, 5051, None, 5000)
        time.sleep(10)
        start_node("node-3", 3, 5002, 5052, None, 5000)
        time.sleep(10)
        res = oai_complete(6000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res["choices"][0]["message"]["content"] + "\"")
        self.assertTrue(len(res["choices"]) > 0)


    def test_reconnect(self):
        start_node("node-1", 1, 5000, 5050, 6000)
        time.sleep(10)
        node2 = start_node("node-2", 1, 5001, 5051, None, 5000)
        time.sleep(10)
        start_node("node-3", 3, 5002, 5052, None, 5000)
        time.sleep(10)
        node2.stop()
        time.sleep(10)
        start_node("node-4", 1, 5004, 5054, None, 5000)
        time.sleep(5)

        res = oai_complete(6000, [
            ChatMessage(ChatRole.SYSTEM, "You are a helpful assistant"),
            ChatMessage(ChatRole.USER, "Hello, how are you?")
        ])
        print("\"" + res["choices"][0]["message"]["content"] + "\"")
        self.assertTrue(len(res["choices"]) > 0)


if __name__ == '__main__':
    unittest.main()


import io
import json
import base64
import ctypes
from uuid import UUID
from hashlib import sha256
from threading import Thread

import torch

def uuid_to_bytes(uid: str) -> bytes:
    return UUID(hex=uid).bytes

def bytes_to_uuid(b: bytes) -> str:
    return str(UUID(bytes=b))

def int_to_bytes(i: int) -> bytes:
    return i.to_bytes(4, 'little', signed=False)

def bytes_to_int(b: bytes) -> int:
    return int.from_bytes(b, 'little', signed=False)

def tensor_to_bytes(t: torch.Tensor) -> bytes:
    bts = io.BytesIO()
    torch.save(t, bts)
    return bts.getvalue()

def bytes_to_tensor(b: bytes) -> torch.Tensor:
    if b == b'':
        return None
    return torch.load(io.BytesIO(b))

def get_tensor_byte_string(t: torch.Tensor) -> str:
    bts = tensor_to_bytes(t)
    return base64.b64encode(bts).decode('utf-8')

def get_hash(data: str) -> str:
    hash = sha256(data.encode())
    return hash.hexdigest()

def get_dict_hash(data: dict):
    return get_hash(json.dumps(data))

def size_of_tensor(t: torch.Tensor):
    return t.element_size() * t.nelement()

def tensor_hash(t: torch.Tensor) -> str:
    return get_hash(get_tensor_byte_string(t))

def stop_thread(thread: Thread):
    thread_id = thread.ident
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id,
            ctypes.py_object(SystemExit))
    if res > 1:
        ctypes.pythonapi.PyThreadState_SetAsyncExc(thread_id, 0)
        print('Exception raise failure')



from time import sleep
from threading import Thread
from typing import Callable, Optional, List
from distributed_state_network import DSNode

from language_pipes.job_manager.pipe import Pipe
from language_pipes.job_manager.job import Job
from language_pipes.job_manager.enums import JobStatus, ComputeStep
from language_pipes.handlers.job import JobServer
from language_pipes.util import stop_thread
from language_pipes.config.processor import ProcessorConfig

class JobReceiver:
    port: int
    public_key_file: str
    private_key_file: str
    router: DSNode
    pending_jobs: List[Job]
    get_pipe: Callable[[str], Optional[Pipe]]
    restart_job: Callable[[Job], None]

    def __init__(
            self, 
            config: ProcessorConfig,
            router: DSNode,
            get_pipe: Callable[[str], Optional[Pipe]],
            restart_job: Callable[[Job], None]
    ):
        self.router = router
        self.get_pipe = get_pipe
        self.restart_job = restart_job
        self.pending_jobs = []

        public_key = router.cert_manager.public_path(router.config.node_id)
        if public_key is None:
            msg = f"Could not find public key for self"
            router.logger.exception(msg)
            raise Exception(msg)

        thread, httpd = JobServer.start(config.https, public_key, config.job_port, self.router, self.receive_data)
        self.thread = thread
        self.httpd = httpd
        Thread(target=self.job_runner, args=()).start()
        router.logger.info(f"Started Job Receiver on port {config.job_port}")

    def job_runner(self):
        while True:
            if self.router.shutting_down:
                return
            if len(self.pending_jobs) == 0:
                sleep(0.1)
                continue
            
            job = self.pending_jobs[-1]
            pipe = self.get_pipe(job.pipe_id)
            if pipe is None or not pipe.is_complete():
                self.restart_job(job)
                return
            
            if job.current_step == ComputeStep.TOKENIZE and job.from_router_id != job.router_id:
                return
            
            if job.current_step != ComputeStep.TOKENIZE and job.status != JobStatus.COMPLETED:
                if job.from_router_id not in pipe.peers():
                    return
            self.pending_jobs.pop()
            pipe.process_job(job)

    def receive_data(self, data: bytes):
        job = Job.from_bytes(data)
        job_certificate = self.router.cred_manager.read_public(job.from_router_id)
        if not job.verify_signature(job_certificate):
            return

        self.pending_jobs.insert(0, job)

    def stop(self):
        self.httpd.stop()
        stop_thread(self.thread)
